package ru.other.company.card.service.provider;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeMap;
import javax.annotation.PostConstruct;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import ru.crystals.pos.api.card.CardEntity;
import ru.crystals.pos.api.plugin.CardPlugin;
import ru.crystals.pos.api.plugin.card.BonusOperationResponse;
import ru.crystals.pos.api.plugin.card.BonusOperationStatus;
import ru.crystals.pos.api.plugin.card.BonusWriteOffOperationResponse;
import ru.crystals.pos.api.plugin.card.CardInfo;
import ru.crystals.pos.api.plugin.card.CardSearchResponse;
import ru.crystals.pos.api.plugin.card.CardSearchResponseStatus;
import ru.crystals.pos.spi.IntegrationProperties;
import ru.crystals.pos.spi.ResBundle;
import ru.crystals.pos.spi.annotation.Inject;
import ru.crystals.pos.spi.annotation.POSPlugin;
import ru.crystals.pos.spi.plugin.card.CardSearchCallback;
import ru.crystals.pos.spi.plugin.card.CardSearchEventSource;
import ru.crystals.pos.spi.plugin.card.CardSearchRequest;
import ru.crystals.pos.spi.receipt.Card;
import ru.crystals.pos.spi.receipt.CardHolder;
import ru.crystals.pos.spi.receipt.Receipt;
import ru.crystals.pos.spi.ui.UIForms;

/**
 * Демонстрационный плагин карточного процессинга, использующий Set API.<br>
 * Плагины карточных процессингов используются для операций бонусными картами, купонами или подарочными картами.
 * Это поиск карты или купона, зачисление бонусов, отмена зачисления бонусов или иная другая операция разработчик плагина позволит себе написать.
 */
@POSPlugin(id = Constants.PLUGIN_NAME)
public class ExampleCardPlugin implements CardPlugin {
    /**
     * Через этот интерфейс мы будем получать настройки плагина.<br>
     * Сами настройки плагина хранятся в БД кассы, куда они поступают с сервера при регистрации и настройке плагина.
     * Какие именно настройки можно получить описывается в metainf.xml плагина, секции &lt;Options&gt;.
     */
    @Inject
    private IntegrationProperties properties;

    /**
     * Логгер, посредством которого будем логировать действия плагина.<br>
     * Его следует именно инжектить, пометя аннотацией {@link ru.crystals.pos.spi.Injectable}, а не создавать при помощи,
     * например {@link org.slf4j.LoggerFactory}.
     */
    @Inject
    private Logger log;

    /**
     * Интерфейс доступа к визуальным формам, доступным для показа плагинам пользователю.<br>
     * Вызов методов этого класса приводит к отображению на экране кассы какого-либо визуального окна.
     */
    @Inject
    private UIForms ui;

    /**
     * Через этот интерфейс будем получать доступ к ресурсам локализации.<br>
     * Это те самые файлы strings_ru.xml, strings_en.xml, которые должны находиться в ресурсах jar с плагином.
     */
    @Inject
    private ResBundle res;

    /**
     * Примерная реализация плагина карт не вызывает внешних сервисов для получения информации по карте,
     * она хранит эту информацию локально. Данное поле - интерфейс доступа к хранилищу карт плагина.
     */
    private CardStorage cardStorage = new CardStorage();

    // Вызов метода, аннотированного PostConstruct доступен только для касс версией от 10.2.47.0 и выше.
    @SuppressWarnings("unused")
    @PostConstruct
    void init() {
        log.info("Plugin {} loaded", Constants.PLUGIN_NAME);
    }

    @Override
    public CardSearchResponse searchCard(CardSearchRequest request) {
        // *
        // * Внимание! Показывать визуальные формы здесь запрещено, вызов любых методов UIForms
        // * не будет иметь никакого эффекта.
        // *
        log.debug("entering getCardInfo({}).", request);
        CardSearchResponse result;
        long stopWatch = System.currentTimeMillis();

        // сначала определим, сможем ли мы этот запрос вообще обработать:
        String cardNo = extractCardNo(request);
        if (cardNo == null || cardNo.length() == 0) {
            // по таким параметрам поиска карту мы не можем идентифицировать:
            result = new CardSearchResponse(CardSearchResponseStatus.UNKNOWN_CARD);
            log.debug("leaving getCardInfo(CardInfoSearchParams). Unable to identify card by search params: {}", request);
            return result;
        }

        // номер карты для поиска известен. Но. может, это заведомо не карта нашего процессинга?
        if (!isCard(cardNo)) {
            // нет. в нашем процессинге карт с таким префиксом (такой длиной) номера нету. Даже в процессинг обращаться не надо
            result = new CardSearchResponse(CardSearchResponseStatus.UNKNOWN_CARD);
            log.debug("leaving getCardInfo(CardInfoSearchParams). The card [{}] does not exist in the processing for sure", cardNo);
            return result;
        }

        // надо обращаться в процессинг. эмулируем задержку:
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
            log.error("interrupted!", e);
        }

        // Эмулируем поиск карты в процессинге:
        result = cardStorage.findCardByNumber(cardNo);
        if(result == null) {
            result = new CardSearchResponse(CardSearchResponseStatus.CARD_NOT_FOUND);
        }

        log.debug("leaving getCardInfo(CardInfoSearchParams). The result is: {}; it took {} [ms] to process the request",
            result, System.currentTimeMillis() - stopWatch);

        return result;
    }

    @Override
    public void searchCardByMobileNumber(String mobileNumber, CardSearchCallback callback) {
        // При поиске карты по номеру телефона допустимо вызывать методы UIForms и управлять
        // визуальными формами.

        // Поиск карты по номеру телефона и обращение в внешний процессинг может быть длительной операцией.
        // Длительные операции могут вызывать панику у кассира на кассе - он начинает воображать, будто касса зависла.
        // Поэтому на время поиска карты покажем кассиру таймер с обратным отсчетом, показывающим, как долго, в худшем случае,
        // придётся выжидать кассиру.
        ui.showTimingOutForm(res.getString("card.search.wait"), 5000, () -> {
            // Мы не справились до таймаута, чтобы не пугать пользователя остановившимся на нуле таймером
            // обратного отсчета, заменим его на сообщение.
            // Следует отметить, натыкание на таймаут до завершения длительной операции не является
            // нормальной ситуацией и этого следует избегать.
            ui.showSpinnerForm(res.getString("wait.some.more"));
        });
        // Симуляция выполнения длительной операции
        // Обратите внимание, велична ожидания здесь выставлена меньше, чем стартовое значение
        // счетчика обратного отсчета на форме ожидания мы отобразили ранее.
        try {
            Thread.sleep(4500);
        } catch (InterruptedException iex) {
            log.error("", iex);
        }
        CardSearchResponse response = cardStorage.findCardByMobileNumber(mobileNumber);
        if(response == null) {
            callback.onCardSearchCompleted(new CardSearchResponse(CardSearchResponseStatus.CARD_NOT_FOUND));
        } else {
            callback.onCardSearchCompleted(response);
        }
    }

    @Override
    public BonusWriteOffOperationResponse writeOff(Card bonusCard, BigDecimal bonusesToWriteOff, Receipt receipt) {
        /*
         * Внимание! Отображать ui здесь нельзя и вызов методов UIForms не будет иметь эффекта.
         */
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
            // NOP
        }
        BonusWriteOffOperationResponse response = new BonusWriteOffOperationResponse();
        response.setOperatorMessage("Списано успешно");
        response.setResponseStatus(BonusOperationStatus.OK);
        response.setTxId("123-456-789");
        return response;
    }

    @Override
    public BonusOperationResponse rollback(String txId) {
        /*
         * Внимание! Отображать ui здесь нельзя и вызов методов UIForms не будет иметь эффекта.
         */
        BonusOperationResponse response = new BonusOperationResponse();
        response.setResponseStatus(BonusOperationStatus.OK);
        response.setTxId("987-654-321");
        return response;
    }

    /**
     * Вернет {@code true}, если возможно(!) указанный номер принадлежит какой-либо карте/купону
     * @param number номер, что надо проверить
     * @return {@code false}, если аргумент невалиден
     */
    private boolean isCard(String number) {
        // либо карта, либо купон:
        return isDiscountCard(number) || isCoupon(number);
    }

    /**
     * Вернет {@code true}, если возможно(!) указанный номер принадлежит какому-либо купону
     * @param number номер, что надо проверить
     * @return {@code false}, если аргумент невалиден
     */
    private boolean isCoupon(String number) {
        if (number == null) {
            return false;
        }
        // если не подойдет по префиксу либо по длине, то вернет false
        Set<String> prefixes = getCouponNoPrefixes();
        if (!meetsPrefixesCriteries(number, prefixes)) {
            return false;
        }
        return properties.getPluginProperties().getInt(Constants.COUPON_NO_LENGTH, Constants.COUPON_NUMBER_LENGTH_DEFAULT) == number.length();
    }

    /**
     * Вернет {@code true}, если возможно(!) указанный номер принадлежит какой-либо дисконтной карте.
     * @param number номер, что надо проверить
     * @return {@code false}, если аргумент невалиден
     */
    private boolean isDiscountCard(String number) {
        if (number == null) {
            return false;
        }
        // если не подойдет по префиксу либо по длине, то вернет false
        Set<String> prefixes = getCardNoPrefixes();
        if (!meetsPrefixesCriteries(number, prefixes)) {
            return false;
        }
        return properties.getPluginProperties().getInt(Constants.CARD_NO_LENGTH, Constants.CARD_NUMBER_LENGTH_DEFAULT) == number.length();
    }

    /**
     * Вернет {@code true}, если указанный номер начинается с одного из указанных префиксов или если список префиксов пуст.
     * @param number номер. что надо протестировать
     * @param prefixes список префиксов
     * @return {@code false}, если данный номер не валиден, либо не удовлетворяет ни одному из префиксов
     */
    private boolean meetsPrefixesCriteries(String number, Set<String> prefixes) {
        if (number == null) {
            return false;
        }
        if (prefixes == null || prefixes.isEmpty()) {
            return true;
        }
        for (String prefix : prefixes) {
            if (prefix == null) {
                continue;
            }
            if (number.startsWith(prefix)) {
                return true;
            }
        }
        // подходящего префикса не нашлось
        return false;
    }

    private Set<String> getCardNoPrefixes() {
        Set<String> cardNoPrefixes = new HashSet<>();
        String commaSeparated = properties.getPluginProperties().get(Constants.CARD_NO_PREFIXES, Constants.CARD_PREFIX_DEFAULT);
        if (commaSeparated != null) {
            cardNoPrefixes.addAll(Arrays.asList(commaSeparated.split(Constants.PREFIX_DELIMITER)));
        }
        return cardNoPrefixes;
    }

    private Set<String> getCouponNoPrefixes() {
        Set<String> couponNoPrefixes = new HashSet<>();
        String commaSeparated = properties.getPluginProperties().get(Constants.COUPON_NO_PREFIXES, Constants.COUPON_PREFIX_DEFAULT);
        if (commaSeparated != null) {
            couponNoPrefixes.addAll(Arrays.asList(commaSeparated.split(Constants.PREFIX_DELIMITER)));
        }
        return couponNoPrefixes;
    }

    /**
     * Из указанных параметров поиска карты выясним номер карты, что надо искать.
     * <p>
     * NOTE: считаем, что наш процессинг может только по номеру карты вести поиск.
     * @param request параметры поиска
     * @return {@code null}, если среди указанных параметров поиска не можем определить номер; иначе - вернет номер карты, что надо искать
     */
    private String extractCardNo(CardSearchRequest request) {
        // наша карточная система считает, что номер карты находится на 2й дорожке магнитной полосы, или оттеснен/напечатан цифрами на карте,
        // или нанесен в виде штрих-кода
        switch (request.getEventSource()) {
            case NUMBER:
            case BARCODE:
                return request.getSearchString();
            case MAGNETIC_STRIPE:
                if(request.getAdditionalData() == null || request.getAdditionalData().isEmpty()) {
                    log.error("Invalid search arguments in request \"{}\": no magnetic strip data provided", request);
                    return null;
                }
                if(request.getAdditionalData().size() < 2) {
                    log.error("Invalid search arguments in request \"{}\": expected 4 magnetic strips, found {}", request, request.getAdditionalData().size());
                    return null;
                }
                // Нулевому элементу в списке соответствует первая дорожка магнитной полосы.
                return request.getAdditionalData().get(1);
            default:
                // других способов идентификации карты наш процессинг не знает:
                return null;
        }
    }

    @Override
    public CardInfo getCardInfo(CardSearchRequest request) {
        String cardNumber = CardSearchEventSource.MAGNETIC_STRIPE.equals(request.getEventSource()) ? request.getAdditionalData().get(1) : request.getSearchString();
        CardSearchResponse response = cardStorage.findCardByNumber(cardNumber);
        if(response == null || !response.getResponseStatus().equals(CardSearchResponseStatus.OK)) {
            return new CardInfo(CardSearchResponseStatus.UNKNOWN_CARD, null);
        }

        CardEntity cardEntity = response.getCard();
        CardHolder cardHolder = cardEntity.getCardHolder();
        String maskedHolderFullName = "";
        if(cardHolder != null) {
            maskedHolderFullName =
                    (StringUtils.isNotEmpty(cardHolder.getLastName()) ? cardHolder.getLastName().substring(0, 1) + ". " : "") +
                    (StringUtils.isNotEmpty(cardHolder.getFirstName()) ? cardHolder.getFirstName() + " " : "") +
                    (StringUtils.isNotEmpty(cardHolder.getMiddleName()) ? cardHolder.getMiddleName() : "");
        }

        TreeMap<String, String> cardInfoMap = new TreeMap<>();
        cardInfoMap.put("Покупатель", maskedHolderFullName);
        cardInfoMap.put("Дата регистрации", "31.08.2018");
        cardInfoMap.put("Статус карты", cardEntity.getCardStatus().name());
        cardInfoMap.put("Номер телефона", "+7 (123) 456-78-90");
        cardInfoMap.put("Баланс бонусов", cardEntity.getBonusBalance() != null ? String.valueOf(cardEntity.getBonusBalance().getBalance()) : "0.00");

        CardInfo cardInfo = new CardInfo(CardSearchResponseStatus.OK, cardInfoMap);
        cardInfo.setTitle("Карта тестового процессинга");
        return cardInfo;
    }
}
