package foo.validation.plugin;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.slf4j.Logger;
import ru.crystals.pos.api.plugin.ExciseValidationPlugin;
import ru.crystals.pos.api.plugin.excise.validation.ExciseValidationResponse;
import ru.crystals.pos.spi.IntegrationProperties;
import ru.crystals.pos.spi.POSInfo;
import ru.crystals.pos.spi.ResBundle;
import ru.crystals.pos.spi.annotation.Inject;
import ru.crystals.pos.spi.annotation.POSPlugin;
import ru.crystals.pos.spi.feedback.Feedback;
import ru.crystals.pos.spi.plugin.excise.validation.ExciseValidationRequest;
import ru.crystals.pos.spi.receipt.Receipt;

/**
 * Пример плагина валидации акцизных марок товаров
 */
@POSPlugin(id = "example.excise.validation.plugin.uniq.id")
public class FooValidationPlugin implements ExciseValidationPlugin {

    /**
     * Название настройки URL процессинга
     */
    private static final String URL = "foo.service.url";

    /**
     * Название настройки таймаута подключения к процессингу
     */
    private static final String CONN_TIMEOUT = "connection.timeout";

    @Inject
    POSInfo posInfo; // параметры кассы

    @Inject
    private IntegrationProperties properties; // настройки интеграции и плагина

    @Inject
    private Logger log; // логирование

    @Inject
    private ResBundle res; // локализованные ресурсы

    @Override
    public ExciseValidationResponse validateExciseForSale(ExciseValidationRequest request) {
        return validate(request, true);
    }

    @Override
    public ExciseValidationResponse validateExciseForRefund(ExciseValidationRequest request) {
        return validate(request, false);
    }

    private ExciseValidationResponse validate(ExciseValidationRequest request, boolean isSale) {
        log.trace("entering validate(ExciseValidationRequest). Excsie: {}, isSale: {}", request.getExcise(), isSale);

        // читаем настройки соединения с процессингом
        String url = properties.getServiceProperties().get(URL);
        int connTimeout = properties.getServiceProperties().getInt(CONN_TIMEOUT, 10000);

        // формируем запрос в процессинг (преобразуем чек в понятный ему формат + возможно, добавляем информацию об окружении (номер магазина, номер кассы))
        // делаем запрос в процессинг и получаем от него результат проверки.
        try {
            Thread.sleep(1000L);
        } catch (InterruptedException e) {
            log.error("interrupted!", e);
            return new ExciseValidationResponse(false, null);
        }
        String message = res.getString("some.localized.message");
        ExciseValidationResponse response = new ExciseValidationResponse(false, message);
        return response;
    }


    @Override
    public Feedback eventReceiptFiscalized(Receipt receipt, boolean isCancelReceipt) {
        if (!isCancelReceipt) { // если это не чек аннулирования
            Feedback feedback;
            try {
                // создадим объект для отправки
                feedback = createFeedback(receipt);
            } catch (JsonProcessingException e) {
                log.error("Error create feedback", e);
                return null;
            }
            try {
                // попытаемся отправить
                trySendFeedBack(feedback.getPayload());
                // если удалось, сообщаем, что больше не надо пытаться отправить
                // т.е. возвращаем null
                return null;
            } catch (Exception e) {
                log.error("Error send feedback", e);
                /**
                 * если при отправке произошла ошибка, то возвращаем кассе объект, который
                 * в дальнейшем будем пытаться отправить.
                 * см. onRepeatSend(ValidationFeedback)
                 */
                return feedback;
            }
        } else {
            return null;
        }
    }

    @Override
    public void onRepeatSend(Feedback feedback) throws Exception {
        trySendFeedBack(feedback.getPayload());
    }

    /**
     * Создание объекта для отправки.<br>
     * В этом примере сериализация в JSON.
     * @param receipt чек
     * @return ValidationFeedback с заполненным payload
     * @throws JsonProcessingException в случае ошибки создания объекта
     */
    private Feedback createFeedback(Receipt receipt) throws JsonProcessingException {
        String json = new ObjectMapper().writeValueAsString(receipt);
        Feedback feedback = new Feedback(json);
        return feedback;
    }

    /**
     * Попытка отправить feedback в процессинг
     * @param payload данные для отправки
     * @throws Exception в случае ошибки отправки
     */
    private void trySendFeedBack(String payload) throws Exception {
        log.trace("trySendFeedBack payload: {}", payload);
        String url = properties.getServiceProperties().get(URL);
        int connTimeout = properties.getServiceProperties().getInt(CONN_TIMEOUT, 10000);
        // здесь происходит отправка данных в процессинг
    }

}
