package ru.crystals.setapi.plugins;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.Spy;
import org.mockito.runners.MockitoJUnitRunner;
import org.mockito.stubbing.Answer;
import org.slf4j.Logger;
import ru.crystals.pos.api.ext.loyal.dto.Slip;
import ru.crystals.pos.api.plugin.goods.NewLineItem;
import ru.crystals.pos.api.ui.listener.DialogListener;
import ru.crystals.pos.api.ui.listener.InputListener;
import ru.crystals.pos.api.ui.listener.InputScanNumberFormListener;
import ru.crystals.pos.spi.IntegrationProperties;
import ru.crystals.pos.spi.PropertiesReader;
import ru.crystals.pos.spi.ResBundle;
import ru.crystals.pos.spi.equipment.CustomerDisplay;
import ru.crystals.pos.spi.equipment.SetApiPrinter;
import ru.crystals.pos.spi.plugin.goods.AddForSaleCallback;
import ru.crystals.pos.spi.plugin.goods.AddForSaleRequest;
import ru.crystals.pos.spi.plugin.goods.CancelDeniedException;
import ru.crystals.pos.spi.plugin.goods.InvalidLineItemException;
import ru.crystals.pos.spi.receipt.Merchandise;
import ru.crystals.pos.spi.receipt.Receipt;
import ru.crystals.pos.spi.ui.UIForms;
import ru.crystals.pos.spi.ui.input.UIInputForms;

import java.util.Collections;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Matchers.anyInt;
import static org.mockito.Matchers.anyObject;
import static org.mockito.Matchers.anyString;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * Пример написания тестов на класс плагина товаров.
 */
@RunWith(MockitoJUnitRunner.class)
public class GoodsPluginExampleTest {

    @Spy
    @InjectMocks
    private GoodsPluginExample plugin;

    @Mock
    private Logger logger;

    @Mock
    private IntegrationProperties properties;

    @Mock
    private ResBundle res;

    @Mock
    private UIForms uiForms;

    @Mock
    private UIInputForms inputForms;

    @Mock
    private AddForSaleCallback callback;

    @Mock
    private SetApiPrinter printer;

    @Mock
    private CustomerDisplay customerDisplay;

    private static final String TEST_BARCODE = "0123456789";

    @Before
    public void initTest() throws Exception {
        plugin.init();
        PropertiesReader reader = mock(PropertiesReader.class);
        when(res.getString(Mockito.anyString())).thenReturn("");
        doAnswer(args -> {
            Slip s = (Slip)args.getArguments()[0];
            System.out.println(s.getParagraphs());
            return null;
        }).when(printer).print(Mockito.any());
        when(properties.getServiceProperties()).thenReturn(reader);
        when(reader.get(eq("server.url"))).thenReturn("http://u.rl");
        when(reader.getInt(eq("timeoutInMs"))).thenReturn(5000);
        doNothing().when(plugin).emulateRemoteRequest();
    }

    @Test
    public void testFindByBarcode() {
        // Проверяем, что товар находится по нашему штрихкоду.
        Merchandise m = plugin.findByBarcode(TEST_BARCODE);
        assertNotNull(m);
        // Проверяем, что все поля заполнены притом.
        assertEquals(TEST_BARCODE, m.getBarcode());
        assertNotNull(m.getName());
        assertEquals(0.0f, m.getNds(), 0.0f);
        assertNotNull(m.getPrice());
        assertTrue(m.getPrice().longValue() > 0);

        // Проверяем, что товар не находится по другому штрихкоду
        m = plugin.findByBarcode("654321");
        assertNull(m);
    }

    @Test
    public void testAddForSalePositive() throws InvalidLineItemException, CancelDeniedException {
        // given
        when(uiForms.getInputForms()).thenReturn(inputForms);
        doAnswer((Answer<Void>) invocation -> {
            InputListener listener = (InputListener) invocation.getArguments()[5];
            listener.eventInputComplete("+7(921)999-99-99");
            return null;
        }).when(inputForms).showPatternInputForm(anyString(), anyString(), anyString(), anyString(), anyString(), Mockito.anyObject());
        doAnswer(args -> {
            DialogListener dl = (DialogListener)args.getArguments()[1];
            dl.eventButton1pressed();
            return null;
        }).when(uiForms).showDialogForm(Mockito.any(), Mockito.any());
        doAnswer((Answer<Void>)invocation -> {
            InputScanNumberFormListener listener = (InputScanNumberFormListener)invocation.getArguments()[4];
            listener.eventBarcodeScanned("1234679");
            return null;
        }).when(inputForms).showInputScanNumberForm(anyString(), anyString(), anyString(), anyInt(), anyObject());

        Merchandise m = plugin.findByBarcode(TEST_BARCODE);
        assertNotNull(m);
        AddForSaleRequest request = mockAddForSaleRequest(m);
        // when
        plugin.addForSale(request);
        // then
        ArgumentCaptor<NewLineItem> captor = ArgumentCaptor.forClass(NewLineItem.class);
        verify(callback, times(1)).completed(captor.capture(), Mockito.any());
        assertNotNull(captor.getValue());
        assertEquals(TEST_BARCODE, captor.getValue().getBarcode());
        assertTrue(captor.getValue().getQuantity() > 0);
        assertTrue(captor.getValue().getPrice().longValue() > 0);
        assertEquals("my-awesome-value", captor.getValue().getData().get("whatever.goods.extended.attribute"));
        assertEquals("attribute-value", captor.getValue().getData().get("whatever.good2.extended.attribute"));
        assertEquals("+7(921)999-99-99", captor.getValue().getData().get("phone.number"));
        verify(callback, times(0)).notCompleted();
        verify(uiForms, times(2)).showSpinnerForm(anyString());
    }

    @Test
    public void testAddForSaleInputCanceled() throws InvalidLineItemException, CancelDeniedException {
        // given
        when(uiForms.getInputForms()).thenReturn(inputForms);
        doAnswer((Answer<Void>) invocation -> {
            InputListener listener = (InputListener) invocation.getArguments()[5];
            listener.eventCanceled();
            return null;
        }).when(inputForms).showPatternInputForm(anyString(), anyString(), anyString(), anyString(), anyString(), Mockito.anyObject());

        Merchandise m = plugin.findByBarcode(TEST_BARCODE);
        AddForSaleRequest request = mockAddForSaleRequest(m);
        // when
        plugin.addForSale(request);
        // then
        verify(callback, times(1)).notCompleted();
        verify(callback, times(0)).completed(anyObject());
        verify(uiForms, times(0)).showSpinnerForm(anyString());
    }

    private AddForSaleRequest mockAddForSaleRequest(Merchandise merchandise) {
        AddForSaleRequest request = mock(AddForSaleRequest.class);
        Receipt receipt = mock(Receipt.class);
        when(request.getMerchandise()).thenReturn(merchandise);
        when(request.getReceipt()).thenReturn(receipt);
        when(receipt.getLineItems()).thenReturn(Collections.emptyList());
        when(request.getCallback()).thenReturn(callback);
        return request;
    }

    @Test
    public void shouldExtractCorrectPhoneNumber() {
        assertEquals("9871234567", GoodsPluginExample.extractPhoneDigits("+7(987)123-45-67"));
        assertEquals("987123456789", GoodsPluginExample.extractPhoneDigits("+3648(987)123-456-789"));
    }

}