package ru.other.company.loy.service.provider;

import java.math.BigDecimal;
import javax.annotation.PostConstruct;
import org.slf4j.Logger;
import ru.crystals.pos.api.ext.loyal.dto.BonusAccrualResult;
import ru.crystals.pos.api.ext.loyal.dto.Discount;
import ru.crystals.pos.api.ext.loyal.dto.LoyaltyResult;
import ru.crystals.pos.api.ext.loyal.dto.Slip;
import ru.crystals.pos.api.ext.loyal.dto.SlipParagraph;
import ru.crystals.pos.api.ext.loyal.dto.SlipParagraphType;
import ru.crystals.pos.api.ext.loyal.dto.auxiliaries.LoyProviderFeedback;
import ru.crystals.pos.api.ext.loyal.dto.auxiliaries.PreFiscalizationFeedback;
import ru.crystals.pos.api.plugin.LoyaltyPlugin;
import ru.crystals.pos.spi.IntegrationProperties;
import ru.crystals.pos.spi.annotation.Inject;
import ru.crystals.pos.spi.annotation.POSPlugin;
import ru.crystals.pos.spi.currency.ExtCurrencyHandler;
import ru.crystals.pos.spi.receipt.Card;
import ru.crystals.pos.spi.receipt.LineItem;
import ru.crystals.pos.spi.receipt.Receipt;
import ru.crystals.pos.spi.receipt.ReceiptType;

/**
 * Примитивный пример реализации {@link LoyaltyPlugin поставщика "услуг лояльности"}: просто дает скидку в 5% на все позиции в чеке.
 *
 * Внимание! В плагинах лояльности недопустимо манипулировать UI кассы, поэтому вызовы методов класса UIForms здесь запрещены.
 * @author aperevozchikov
 */
@POSPlugin (id = ExampleLoyProvider.NAME)
public class ExampleLoyProvider implements LoyaltyPlugin {
    /**
     * код этого поставщика услуг лояльности
     */
    public static final String NAME = "ext.loy.processing.example";

    /**
     * название настройки, что хранит URL для подключения к процессингу
     */
    private static final String URL = "primary.calc.url";

    /**
     * название настройки, что хранит таймаут подключения к процессингу
     */
    private static final String CONN_TIMEOUT = "connection.timeout";

    /**
     * название настройки, что хранит время ожидания отклика от процессинга
     */
    private static final String READ_TIMEOUT = "read.timeout";

    /**
     * Идентификатор единственной Рекламной Акции в процессинге
     */
    private static final String REGULAR_DISCOUNT_ID = "plugin-discount";

    private static final String BONUS_DISCOUNT_ID = "bonus-discount";

    /**
     * Доля оригинальной стоимости позиции, что дадим в виде скидки
     */
    private static final BigDecimal DISCOUNT_SHARE = new BigDecimal(0.05);

    /**
     * Идентификатор карты, которая считается "родственной" плагину лояльности.
     * Это идентификатор того карточного плагина, который работает в паре с этим плагином лояльности
     * и выполняет поиск, списание и отмену списания по картам.
     * Плагину лояльности нужно знать "свои" карты, чтобы ошибочно не совершить операции с картами не своего процессинга.
     */
    private static final String MY_CARD_ID = "example.card.provider";
    /**
     * Округлятор денег. Для операций с валютой (округление, вычисление стоимости) следует использовать его.
     */
    @Inject
    private ExtCurrencyHandler currencyHandler;

    /**
     * При помощи этого класса осуществляется доступ к настройкам плагина и внешней системы он подчинен.
     * Не требуется создавать экземпляр этого класса, он будет создан загрузчиком плагина автоматически при.
     */
    @Inject
    private IntegrationProperties properties;

    /**
     * Логгер, который будет собирать весь лог плагина. Его не следует создавать вручную, ровно как и использовать
     * какой-либо иной логгер.
     */
    @Inject
    private Logger log;

    // Вызов метода, аннотированного PostConstruct доступен только для касс версией от 10.2.47.0 и выше.
    @SuppressWarnings("unused")
    @PostConstruct
    void init() {
        log.info("Plugin {} loaded", this.getClass());
    }

    @Override
    public LoyProviderFeedback eventReceiptFiscalized(Receipt receipt, LoyaltyResult loyResults) {
        // Наш плагин интересуют только чеки продажи
        if(receipt.getType() != ReceiptType.SALE) {
            return null;
        }
        LoyProviderFeedback feedback = new LoyProviderFeedback();
        feedback.setPayload("Plugin payload goes here");
        return feedback;
    }

    @Override
    public LoyProviderFeedback onSendFeedback(LoyProviderFeedback feedback) {
        log.trace("onSendFeedback({}) called", feedback);
        return null;
    }

    @Override
    public LoyaltyResult doDiscount(Receipt receipt) {
        log.trace("entering process(Receipt). The receipt-no is: {}", receipt == null ? null : receipt.getNumber());
        LoyaltyResult result;
        long stopWatch = System.currentTimeMillis();
        if (receipt == null) {
            log.error("leaving process(Receipt): the argument is NULL");
            return null;
        }
        // читаем настройки соединения с процессингом лояльности
        String url = properties.getServiceProperties().get(URL);
        int connTimeout = properties.getServiceProperties().getInt(CONN_TIMEOUT, 10000);
        int readTimeout = properties.getServiceProperties().getInt(READ_TIMEOUT, 10000);

        // формируем запрос в процессинг (преобразуем чек в понятный ему формат + возможно, добавляем информацию об окружении (номер магазина, номер кассы))
        //  в примере это требуется

        // делаем запрос в процессинг и получаем от него результат расчета скидок.
        try {
            Thread.sleep(1000L);
        } catch (InterruptedException e) {
            log.error("interrupted!", e);
            return null;
        }

        // преобразуем отклик процессинга в LoyaltyResult (возможно, применяя некий пост-процессинг (коррекцию размеров скидок, данных процессингом))
        //  в примере это требуется - просто дадим скидку в 5% на все позиции
        result = new LoyaltyResult();
        // Просто для примера - если есть "наша" карта, позволим списать с неё 3 бонуса как скидку
        String cardNumber = null;
        for(Card card : receipt.getCards()) {
            if(!MY_CARD_ID.equals(card.getProcessingId())) {
                log.debug("Card \"{}\" (\"{}\") does not belong to \"{}\". Skipping it.", card.getCardNumber(), card.getProcessingId(), ExampleLoyProvider.NAME);
                continue;
            }
            cardNumber = card.getCardNumber();
            if(card.getBonusBalance() != null && card.getBonusBalance().getSumToChargeOff() != null &&
                    card.getBonusBalance().getSumToChargeOff().compareTo(BigDecimal.ZERO) > 0) {
                // Если мы попали сюда, с карты списывались бонусы как скидка.
                // Сейчас плагины умеют предоставлять скидки только на позиции, а не на весь чек, поэтому размажем величину скидки по позициям.
                // Для упрощения, предоставим скидку на первую позицию, слежение за МРЦ и отрицательной стоимостью опущено.
                result.getDiscounts().add(new Discount(1, card.getCardNumber(), card.getBonusBalance().getSumToChargeOff(), BONUS_DISCOUNT_ID));
            } else {
                // С этой карты не пытались списать бонусы ещё, заполним максимально разрешенную величину к списанию.
                // Это даст возможность списать их как скидка и учесть при следующем расчете (тогда мы попадём в верхний блок вместо сюда).
                // Для примера, просто заполним фиксированной.
                result.getBonusWriteOffsLimits().put(card.getCardNumber(), BigDecimal.valueOf(3));
            }
        }
        for (LineItem position : receipt.getLineItems()) {
            if (position == null || !position.isDiscountable()) {
                // на эту позицию нельзя давать скидок
                continue;
            }

            // оригинальная стоимость позиции
            BigDecimal originalSum = position.getSum().add(position.getDiscountValue());

            // минимально допустимая стоимость позиции
            BigDecimal minSum = currencyHandler.getPositionSum(position.getMinPrice(), position.getQuantity());

            // максимально допустимый размер скидки, что еще можно дать на эту позицию
            BigDecimal maxDiscount = position.getSum().subtract(minSum);
            if (maxDiscount.compareTo(BigDecimal.ZERO) <= 0) {
                // на эту позицию больше нельзя давать скидок
                continue;
            }

            // скидка в 5% на эту позицию:
            BigDecimal discountValue = currencyHandler.round(originalSum.multiply(DISCOUNT_SHARE));
            if (discountValue.compareTo(maxDiscount) >= 0) {
                // все 5% не можем скинуть из-за ограничений минимальной цены. Отдадим сколько сможем:
                discountValue = maxDiscount;
            }

            Discount discount = new Discount(position.getNumber(), null, discountValue, REGULAR_DISCOUNT_ID);
            result.getDiscounts().add(discount);
        }

        // Добавим несколько печатных слипов к расчету скидок
        addPrintSlips(result);
        // Добавим сообщение кассру, которое отображается после расчета скидок
        result.getMessagesToCashier().add("Выдайте клиету в подарок сдачу.");

        // Эти атрибуты будет сохранён в чеке и будут выгружаться в ERP. Атрибуты будут доступны плагину с момента добавления и на протяжении
        // всей жизни чека.
        // Хорошей практикой является именовать атрибуты, добавляя идентификатор своего плагина в качестве префикса.
        // Таким образом, будет меньше шансов, что два разных плагина непредусмотрено станут использовать один и тот же атрибут, перетирая его.
        result.getExtendedReceiptAttributesMap().put("ext.loy.processing.example.external.receipt.id", "123469877");
        result.getExtendedReceiptAttributesMap().put("ext.loy.processing.example.some.other.field", "whatever");

        // Мы можем уведомить кассу о том, что плагин лояльности начислил на карту бонусные баллы.
        // Это необходимо, чтобы отобразить событие начисление бонусов на карту на сервере в Опердне или выгрузке в ERP.
        // При отмене расчета скидок уведомление будет удалено кассой автоматически.
        if(cardNumber != null && receipt.getType() == ReceiptType.SALE) {
            // Уведомляем кассу о начислении десяти бонусов на карту.
            result.getBonusAccrualResults().put(cardNumber, new BonusAccrualResult("123456-8789", BigDecimal.valueOf(10.0)));
        }
        log.trace("leaving process(Receipt). The result is: {}; it took {} [ms] to process the request", result, System.currentTimeMillis() - stopWatch);
        return result;
    }


    /**
     * Добавляет печатные слипы к результату расчета скидок в внешней системе.
     * @param result результат расчета скидок требуется добавить печатные слипы которому.
     * @return тот же экземпляр результата расчета скидок, но с добавленными печатными слипами.
     */
    private LoyaltyResult addPrintSlips(LoyaltyResult result) {
        Slip slip1 = new Slip();
        slip1.getParagraphs().add(
                new SlipParagraph(SlipParagraphType.TEXT, "Реклама на чеке:\nКупите пряник малиновый сладкий и получите ничего в подарок!")
        );
        result.getPrintDocuments().add(slip1);

        Slip slip2 = new Slip();
        slip2.getParagraphs().add(
                new SlipParagraph(SlipParagraphType.TEXT, "Спасибо, что воспользовались нашим плагином лояльности.\nЭто второй печатный слип")
        );
        // Это строка в слипе будет распечатана как баркод.
        slip2.getParagraphs().add(new SlipParagraph(SlipParagraphType.BARCODE, "123469879792545454"));
        result.getPrintDocuments().add(slip2);
        return result;
    }

    @Override
    public PreFiscalizationFeedback eventBeforeReceiptFiscalized(Receipt receipt, LoyaltyResult loyResults) {
        Slip slip = new Slip();
        slip.getParagraphs().add(new SlipParagraph(SlipParagraphType.TEXT, "Пример слипа при событии предфискализации чека"));
        slip.getParagraphs().add(new SlipParagraph(SlipParagraphType.TEXT, "Примененная скидка: 0р."));

        PreFiscalizationFeedback preFiscalizationFeedback = new PreFiscalizationFeedback();
        preFiscalizationFeedback.getSlips().add(slip);

        if(receipt.getType() != ReceiptType.SALE) {
            return preFiscalizationFeedback;
        }
        // До фискализации всё ещё допустимо начислить бонусов на карту.
        for(Card card : receipt.getCards()) {
            if(!MY_CARD_ID.equals(card.getProcessingId())) {
                continue;
            }
            preFiscalizationFeedback.getBonusAccrualResults()
                    .put(card.getCardNumber(), new BonusAccrualResult("978-789-546", BigDecimal.valueOf(13.0)));
        }

        return preFiscalizationFeedback;
    }
}
