package ru.crystals.pos.api.plugin;

import ru.crystals.pos.api.ext.loyal.dto.DiscountCancellationResult;
import ru.crystals.pos.api.ext.loyal.dto.LoyaltyResult;
import ru.crystals.pos.api.ext.loyal.dto.auxiliaries.LoyProviderFeedback;
import ru.crystals.pos.api.ext.loyal.dto.auxiliaries.PreFiscalizationFeedback;
import ru.crystals.pos.spi.receipt.Receipt;

/**
 * Поставщик "услуг лояльности": может дать скидку на чек, добавить купон, рекламный буклет, насчитать бонусы, и проч.
 *
 * @author aperevozchikov
 */
public interface LoyaltyPlugin {

    /**
     * Подсчитает скидки и другие "преференции" лояльности для указанного чека.
     *
     * @param receipt
     *            чек, <em>с уже данными скидками/преференциями другими "поставщиками лояльности"</em>
     * @return {@code null}, если данный "поставщик лояльности" не активен (отключен), либо аргументы невалидны; иначе - вернет результат применения
     * программ лояльности от данного поставщика "услуг лояльности" к данному чеку
     */
    LoyaltyResult doDiscount(Receipt receipt);

    /**
     * Обрабатывает событие "отмена расчета скидок".<br>
     * Событие наступает, например, в случае, если из режима оплат чека переходят обратно в режим добавления товаров (до того, как хотя бы одна оплата будет добавлена)
     * или при аннулировании чека.
     * @param receipt чек, отменить расчет скидок в котором необходимо.
     * @param results результаты расчета скидок по данному чеку
     * @return структуру {@link DiscountCancellationResult}, содержащую результаты отмены расчета скидок в плагине.
     * @since 0.0.26
     */
    default DiscountCancellationResult cancelDiscount(Receipt receipt, LoyaltyResult results) {
        return null;
    }

    /**
     * Оповещение данного поставщика услуг лояльности о факте фискализации чека.
     * <p>
     * Предполагается, что получив это оповещение, поставщик лояльности подготовит данные обратной связи для своего процессингового центра (ПЦ) -
     * чтоб оповестить его (ПЦ) о предоставленных преференциях в чеке.
     *
     * @param receipt       чек, что был фискализирован
     * @param loyResults    преференции, что данный поставщик услуг лояльности предоставил в этом чеке
     * @return {@code null}, если никакого фидбэка для процессинга не требуется
     */
    default LoyProviderFeedback eventReceiptFiscalized(Receipt receipt, LoyaltyResult loyResults) {
        return null;
    }

    /**
     * Событие предфискализации.
     * Вызывается непосредственно перед тем, как чек будет фискализирован.
     *
     * @param receipt - чек, который будет фискализирован
     * @param loyResults    преференции, что данный поставщик услуг лояльности предоставил в этом чеке
     * @return {@code null}, если никакого фидбэка для процессинга не требуется
     * */
    default PreFiscalizationFeedback eventBeforeReceiptFiscalized(Receipt receipt, LoyaltyResult loyResults) {
        return null;
    }

    /**
     * Попытка отправить указанный фидбэк в Процессинговый Центр данного поставщика "услуг лояльности".
     * <p>
     * NOTE: данный метод будет вызываться периодически по мере обнаружения фидбэков, что данный поставщик "услуг лояльности" хотел бы передать в свой ПЦ.
     *
     * @param feedback    представление (XML, JSON-строка, base64, и проч.) фидбэка, что надо отправить.
     *                    NOTE: данный фидбэк был сформирован самим данным плагином лояльности: как результат метода {@link #eventReceiptFiscalized(Receipt, LoyaltyResult)}
     *                    или {@link #onSendFeedback(LoyProviderFeedback)}
     * @return
     * <ol>
     * <li> {@code null}, если более не надо пытаться отправить данный фидбэк в ПЦ (Процессинговый Центр) - по любой причине (либо фидбэк успешно доставлен,
     * либо фидбэк признан "битым" и потому не стоит более пытаться его отправить, т.к. ПЦ его все равно не сможет "понять");
     * <li> фидбэк, что надо попробовать отправить взамен указанного в следующий раз (скорей всего,
     * с той же самой {@link LoyProviderFeedback#getPayload() полезной нагрузкой}, но с увеличенным количеством
     * {@link LoyProviderFeedback#getAttemptsCount()} неудачных попыток отправки})
     * </ol>
     */
    default LoyProviderFeedback onSendFeedback(LoyProviderFeedback feedback) {
        return null;
    }

}
