package ru.crystals.pos.spi.receipt;

import java.math.BigDecimal;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * Представление чека кассы.
 */
public interface Receipt {

    /**
     * Получить общую сумму чека.
     * 
     * @return сумма чека с двумя знаками после запятой, т.е. scale=2
     */
    BigDecimal getSumWithDiscount();
    
    /**
     * Получить оставшуюся сумму к оплате/возврату.
     * <pre>
     *  Сумма чека минус сумма всех уже выполненных в этом чеке оплат/возвратов.
     * </pre>
     * @return сумма с двумя знаками после запятой, т.е. scale=2
     */
    BigDecimal getSurchargeSum();

    /**
     * Получить позиции данного чека
     *
     * @return не {@code null}
     */
    List<LineItem> getLineItems();

    /**
     * Получить номер смены, к которой относится данный чек
     *
     * @return номер смены
     */
    int getShiftNo();

    /**
     * Получить [Прогнозируемый] номер этого чека.
     * <p>
     * NOTE: действительный номер присваивается документу (чеку) после успешной фискализации - до этого момента его порядковый номер можно только <em>прогнозировать</em>.
     *
     * @return номер данного чека
     */
    int getNumber() ;

    /**
     * Получить коллекцию карт, что были применены в данном чеке
     *
     * @return не {@code null}
     */
    Collection<Card> getCards();

    /**
     * Получить коллекцию акцизных бутылок алкогольного товара в данном чеке
     *
     * @return не может быть {@code null}
     */
    Collection<ExciseBottle> getExciseBottles();

    /**
     * Возвращает тип чека.
     * @since 0.0.20
     * @see ReceiptType#SALE
     * @see ReceiptType#REFUND
     * @return тип чека.
     */
    ReceiptType getType();

    /**
     * В случае чека возврата, здесь может находиться оригинальный чек продажи, по которому осуществляется возврат.<br>
     * Следует отметить, не всякие чеки возврата могут иметь чек продажи, например, если осуществляется произвольный возврат
     * (возврат покупки без чека).
     * @see #getType()
     * @see ReceiptType#REFUND
     * @since 0.0.20
     * @return чек возврата или null, если такого нет
     */
    Receipt getSaleReceipt();

    /**
     * Получает мапу атрибутов, которые были добавлены к чеку.<br>
     * Позволяет добавлять к чеку любые дополнительные параметры, которые затем можно использовать где-нибудь в другом месте.
     * Мапа иммутабельна, хранит только строковые типы.
     * @return мапу атрибутов, которые были добавлены к чеку или пустую мапу, если таких нет.
     * @since 0.0.22
     */
    default Map<String, String> getData() {
        return Collections.EMPTY_MAP;
    }

    /**
     * Возвращает дату создания оригинального чека
     * @return дата создания чека
     * @since 0.0.24
     * */
    Date getDateCreate();
}
